#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "filecache.h"
#include "xml.h"

#define DEBUG

cachedfiles::cachedfiles(nsPluginInstance * plugin_instance,
			 const char *url)
{
	files = NULL;
	nfiles = 0;
	maxfiles = 0;
	instance = plugin_instance;

#ifdef DEBUG
	printf("%40s(): New cache for base url '%s'.\n",
	       __PRETTY_FUNCTION__, url);
#endif

	if (setup_cachelist(512) == -1)
		printf("%40s(): Cache list not available.\n",
		       __PRETTY_FUNCTION__);

	base_url = (char *) url;

	pthread_mutex_init(&cache_mutex, NULL);
	pthread_cond_init(&new_file_cond, NULL);
}


cachedfiles::~cachedfiles(void)
{
	char *listfilename = "/var/tmp/smilplugin-cache";
	char *listfilebkname = "/var/tmp/smilplugin-cache.bak";
	FILE *listfilebk;
	int n = 0;

	/* If the backup file exists, we try to wait a few seconds.
	 */
	while (access(listfilebkname, F_OK) == 0) {
#ifdef DEBUG
		printf("%40s(): Cache list locked...\n",
		       __PRETTY_FUNCTION__);
#endif
		if (n > 2) {
			printf("%40s(): Cachelist backupfile left?\n",
			       __PRETTY_FUNCTION__);
			return;
		}
		sleep(1);
		n++;
	}

	/* If there is a cache list file we move it to the safe
	 * name to lock out others.
	 */
	if (access(listfilename, F_OK) == 0)
		if (rename(listfilename, listfilebkname) != 0) {
			printf("%40s(): Can not rename '%s' to '%s'.\n",
			       __PRETTY_FUNCTION__, listfilename,
			       listfilebkname);
			return;
		}

	/* We open the backup file for append write out the list
	 * of cached files and close the file at the end..
	 */
	listfilebk = fopen(listfilebkname, "a");
	if (!listfilebk) {
		printf("%40s(): Can't open '%s' for append: %m\n",
		       __PRETTY_FUNCTION__, listfilebkname);
		return;
	}

	for (n = 0; n < nfiles; ++n) {
		/* Deciding what to cache and what to delete is a complex task.
		 * FIXME: a stat() and timestamp check should be nice.
		 */
		if (!files[n].filename)
			continue;
		if (files[n].state & STATE_FROM_CACHE)
			continue;
		if (!files[n].url ||
		    !(files[n].state & STATE_DOWNLOADED) ||
		    files[n].actualsize < 1024 * 1024) {
#ifdef DEBUG
			printf("%40s(): deleting '%s'\n",
			       __PRETTY_FUNCTION__, files[n].filename);
#endif
			unlink(files[n].filename);
			continue;
		}
		fprintf(listfilebk, "%s\t%s\n", files[n].url,
			files[n].filename);
	}

	fclose(listfilebk);

	/* Sorry, we left a backup file. This is not a nice move.
	 */
	if (rename(listfilebkname, listfilename) != 0)
		printf("%40s(): Can not rename '%s' to '%s'.\n",
		       __PRETTY_FUNCTION__, listfilebkname, listfilename);

	/*
	 * Destroyink pthread resources. I'm not soure why I need destroy them
	 * but I make it.
	 */
	pthread_mutex_destroy(&cache_mutex);
	pthread_cond_destroy(&new_file_cond);
}


int cachedfiles::setup_cachelist(int n)
{
	files = (file *) NPN_MemAlloc(sizeof(file) * n);
	if (!files)
		return -1;
#ifdef DEBUG
	printf("%40s(): Cache list created for %d files.\n",
	       __PRETTY_FUNCTION__, n);
#endif

	nfiles = 0;
	maxfiles = n;
	return 0;
}


/*
 * This function will search the hard drive cache to search an already downloaded
 * version of the url file.
 */
int cachedfiles::find_in_cache(file * thisfile)
{
	char *listfilename = "/var/tmp/smilplugin-cache";
	FILE *listfile;
	char *buffer = NULL;
	size_t linelength = 0;
	char *filename = NULL;
	int found = 0;

	if (!thisfile || !thisfile->url)
		return 0;

	listfile = fopen(listfilename, "r");
	if (!listfile)
		return 0;

	while (getline(&buffer, &linelength, listfile) > 0) {
		if (filename = strchr(buffer, '\t')) {
			*filename = '\0';
			if (strcmp(thisfile->url, buffer) == 0 &&
			    filename != NULL) {
				++filename;
				filename[strlen(filename) - 1] = '\0';
				thisfile->filename = strdup(filename);
				thisfile->state |= STATE_FROM_CACHE;
				found = 1;
				break;
			}
		}
	}

	if (buffer)
		free(buffer);

	if (found) {
#ifdef DEBUG
		printf("%40s(): Cache hit for file '%s'\n",
		       __PRETTY_FUNCTION__, thisfile->filename);
#endif
		file_available(thisfile);
	}

	return found;
}


const char *cachedfiles::findfilename(const char *url)
{
	int n;
	size_t size = strlen(url) + strlen(base_url) + 1;
	char *canonical_url = (char *) alloca(size);

	canonicalize_url(url, canonical_url, size);

	for (n = 0; n < nfiles; ++n) {
		if (strcmp(files[n].url, canonical_url) == 0)
			return files[n].filename;
	}

#ifdef DEBUG
	printf("%40s(): file not found in %d for '%s'\n",
	       __PRETTY_FUNCTION__, nfiles, canonical_url);
#endif

	return NULL;
}


file *cachedfiles::findfile(const char *url)
{
	int n;
	size_t size = strlen(url) + strlen(base_url) + 1;
	char *canonical_url = (char *) alloca(size);

	canonicalize_url(url, canonical_url, size);

	for (n = 0; n < nfiles; ++n) {
		if (strcmp(files[n].url, canonical_url) == 0)
			return &files[n];
	}

#ifdef DEBUG
	printf("%40s(): file not found in %d for '%s'\n",
	       __PRETTY_FUNCTION__, nfiles, canonical_url);
#endif

	return NULL;
}

/*
 * This function will close the file if it is opened and sets its state 
 * according the process.
 */
int cachedfiles::close_cache_file(file * to_close)
{
	if (!(to_close->state & STATE_OPENED_READ) &&
	    !(to_close->state & STATE_OPENED_READ))
		return -1;

	if (close(to_close->filehandler) == -1)
		printf("%40s(): Error closing '%s': %m\n",
		       __PRETTY_FUNCTION__, to_close->filename);

	to_close->filehandler = 0;
	to_close->state &= ~STATE_OPENED_WRITE;
	to_close->state &= ~STATE_OPENED_READ;

	return 0;
}

/*
 * This function will open the cache file for reading and writing and
 * create a new cache if necessary.
 */
int cachedfiles::open_cache_file(file * to_open, int flags)
{
	int state;
	char *filename_template = NULL;

	if (!to_open->filename && !(flags & O_CREAT)) {
		printf("%40s(): File has no name and no creation wanted.",
		       __PRETTY_FUNCTION__);
		return -1;
	}

	if (!to_open->filename) {
		filename_template = (char *) NPN_MemAlloc(32);
		if (!filename_template) {
			printf("%40s(): No memory.\n",
			       __PRETTY_FUNCTION__);
			return -1;
		}
		strcpy(filename_template, "/var/tmp/smilplugin-XXXXXX");
		to_open->filename = mktemp(filename_template);
		if (!to_open->filename) {
			printf("%40s(): Tempfile creation error: %m",
			       __PRETTY_FUNCTION__);
			return -1;
		}
	}

	/*
	 * If the file is opened we close and reopen.
	 */
	close_cache_file(to_open);

	to_open->filehandler = open(to_open->filename, flags,
				    S_IRWXU | S_IRWXG | S_IRWXO);
	if (!to_open->filehandler) {
		printf("%40s(): Unable to open '%s': %m\n",
		       __PRETTY_FUNCTION__, to_open->filename);
		return -1;
	}

	if (flags & O_RDONLY || flags & O_RDWR)
		to_open->state |= STATE_OPENED_READ;

	if (flags & O_WRONLY || flags & O_RDWR)
		to_open->state |= STATE_OPENED_WRITE;

	return to_open->filehandler;
}


char *cachedfiles::canonicalize_url(const char *url, char *buffer, size_t size){
	const char *tmp;
	char *end;
	
	if (strstr(url, "://")){
		strncpy(buffer, url, size);
		return buffer;
	}
	
	if (strlen(base_url) >= size)
		return NULL;
	
	strncpy(buffer, base_url, size);
	
	tmp = url;
	end = buffer + strlen(buffer) - 1;
	while (!strncmp(tmp, "../", 3)){
		tmp += 3;
		while (end-- != buffer){
			*(end + 1) = '\0';
			if (*end == '/')
				break;
		}
	}
	
	strcat(buffer, tmp);
	return buffer;
}

/*
 * Returns true if the file needed to download, false if already
 * found.
 */
int cachedfiles::newfile(const char *url, const char *mimetype,
			 uint32 size)
{
	size_t tmpsize = strlen(url) + strlen(base_url) + 1;
	char *canonical_url = (char *) alloca(tmpsize);

	canonicalize_url(url, canonical_url, tmpsize);
	
#ifdef DEBUG
	printf("%40s(): Creating new file for '%s'\n", __PRETTY_FUNCTION__,
	       url);
#endif

	if (findfile(canonical_url)) {
		printf("%40s(): This file is already cached.\n",
		       __PRETTY_FUNCTION__);
		return FALSE;
	}

	if (nfiles == maxfiles) {
		printf("%40s(): Too many files in cache.\n",
		       __PRETTY_FUNCTION__);
		return FALSE;	/* FIXME: realloc would be nicer. */
	}

	files[nfiles].url = strdup(canonical_url);
	files[nfiles].filename = NULL;
	files[nfiles].filesize = size;
	files[nfiles].actualsize = 0;
	files[nfiles].state = STATE_INITIALIZED;
	if (mimetype)
		files[nfiles].mimetype = strdup(mimetype);
	else
		files[nfiles].mimetype = NULL;

	nfiles++;

	return (find_in_cache(&files[nfiles - 1]) == 0);
}

/*
 * Returns -1 if some error occured or the number of written bytes
 * if everithing went well.
 */
int cachedfiles::writefile(int32 offset, int32 length, void *buffer,
			   const char *url)
{
	ssize_t written;
	file *thisfile = findfile(url);
	if (!thisfile)
		newfile(url);

	thisfile = findfile(url);
	if (!thisfile)
		return -1;

	if (!(thisfile->state & STATE_OPENED_WRITE))
		if (!open_cache_file(thisfile, O_RDWR | O_CREAT))
			return -1;

#ifdef LOTOFDEBUG 
	printf("%40s(): Writing %d bytes at offset %d to file '%s'\n",
	       __PRETTY_FUNCTION__, length, offset, thisfile->filename);
#endif

	written = pwrite(thisfile->filehandler, buffer, length, offset);
	if (written != length) {
		printf("%40s(): Writing error on '%s': %m\n",
		       __PRETTY_FUNCTION__, thisfile->filename);
		printf("%40s(): Buffer %d of %d wrote\n",
		       __PRETTY_FUNCTION__, written, length);
		return -1;
	}

	thisfile->actualsize = thisfile->actualsize + written;
	if (thisfile->actualsize == thisfile->filesize)
		thisfile->state |= STATE_DOWNLOADED;
#ifdef LOTOFDEBUG 
	printf("%40s(): Wrote %d byte to '%s'\n", __PRETTY_FUNCTION__,
	       written, thisfile->filename);
#endif

	if (thisfile->state & STATE_DOWNLOADED)
		file_available(thisfile);

	return length;
}

/*
 * We call this function when the file become available (download
 * complete or found in cache).
 */
int cachedfiles::file_available(file * thisfile)
{
	/* If a new file become available, we tell it to the threads waiting
	 * for any file.
	 */
	pthread_cond_signal(&new_file_cond); /*FIXME: I don't use this, do I?*/

	return instance->FileDownloaded(thisfile->filename,
					thisfile->mimetype);
}

file *cachedfiles::get_first_file(void){
	current = 0;
	if (nfiles > 0)
		return &files[0];
	else
		return NULL;
}

file *cachedfiles::get_next_file(void){
	current++;
	
	if (nfiles > current)
		return &files[current];
	else
		return NULL;	
}

int cachedfiles::print_cached_file_list(void)
{
	int n;

	printf("----cached file list follows---------------\n");

	for (n = 0; n < nfiles; ++n) {
		printf("%04d: Filename: '%s'\n", n, files[n].filename);
		printf("      url:      '%s'\n", files[n].url);
		printf("      state: %04x size: %12d actualsize:%12d\n",
		       files[n].state, files[n].filesize,
		       files[n].actualsize);

		if (files[n].mimetype)
			printf("      mimetype: '%12s'\n",
			       files[n].mimetype);
		else
			printf("      mimetype unknown");

	}
	printf("-------------end of list-------------------\n");

	return 0;
}


