#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>
#include "smilplugin.h"
#include "filecache.h"
#include "xml.h"
#include "player.h"

#define DEBUG

void *player_thread_entry(void *data);
void *show_thread_entry(void *data);

#define MULTI_WINDOW

player::player(nsPluginInstance *ins)
{
	printf("%40s(): Creating new player.\n", __PRETTY_FUNCTION__);
	instance = ins;
	window = NULL;
	x_window = 0;
	current_media = NULL;
	state = STATE_FREE; 
	current_second = -1;
	control_stream = NULL;
	child = -1;
}

player::~player(void){

	state |= STATE_QUIT;

	if (window){
		XFreeGC(x_display, x_gc);
		XFreeGC(x_display, x_inv_gc);
	}
	
	if (control_stream){
#ifdef DEBUG
		printf("%40s: Stopping mplayer.\n", __PRETTY_FUNCTION__);
		fflush(stdout);
#endif
		fprintf(control_stream, "quit\n");
		fflush(control_stream);
	}
	sleep(1);

	if (child != -1){
#ifdef DEBUG
		printf("%40s: Killing child.\n", __PRETTY_FUNCTION__);
		fflush(stdout);
#endif
		kill(child, 1);
		child = -1;
	}
	
  	while (waitpid(WAIT_ANY, NULL, WNOHANG) > 0)
	;
#ifdef DEBUG
		printf("%40s: Ready.\n", __PRETTY_FUNCTION__);
		fflush(stdout);
#endif
}

/*
 *
 *
 */
void player::draw_status(char *lines[])
{
	int n;

	if (!window || state & STATE_PLAYING)
		return;

	/*
   	 * First we write the name of the program.
	 */
	XFillRectangle(x_display, (Drawable)x_window, 
				x_inv_gc, 0, 0, window->width, 13);
	XDrawString(x_display, (Drawable) x_window, 
			x_gc, 2, 10, PACKAGE_STRING , 
			strlen(PACKAGE_STRING));
	/*
	 * Now the lines of text.
	 */
	n = 0;
	while (lines[n]){
		XFillRectangle(x_display, (Drawable)x_window, 
				x_inv_gc, 0, 32 + x_fontheight * n,
				window->width, x_fontheight);

		XDrawString(x_display, (Drawable) x_window, 
				x_gc, 2, 30 + x_fontheight * (n + 1), 
				lines[n], strlen(lines[n]));
		n++;
	}

	XFlush(x_display);
}


int player::start_play_this(media_instance_descriptor *media)
{
	file *file_in_cache;
	
	printf("%s(): Starting.\n", __PRETTY_FUNCTION__);

	if (state & STATE_QUIT){
#ifdef DEBUG
		printf("%s: Start when destroying player.\n",
				__PRETTY_FUNCTION__);
#endif
		return -1;
	}

	state = STATE_STARTING;

	if (!media || !media->src){
		printf("%40s: Media not found.\n", __PRETTY_FUNCTION__);
		return -1;
	}

	// FIXME: sometimes this function gets media without src.
	//printf("%s(): Start playing '%s'\n", __FUNCTION__, media->src);

	current_media = media;
	
	pthread_mutex_lock(&instance->mutex_filecame);
	file_in_cache = instance->filecache->findfile(media->src); 
	while (!file_in_cache){
		printf("%40s: Not found in cache: '%s'\n", __PRETTY_FUNCTION__, media->src);
		printf("%40s: Going to sleep.\n", __PRETTY_FUNCTION__);
		instance->filecache->print_cached_file_list();
		fflush(stdout);
		state = STATE_WAITING;
		pthread_cond_wait(&instance->cond_filecame, 
				&instance->mutex_filecame);
		file_in_cache = instance->filecache->findfile(media->src); 
		return -1;
	}
	pthread_mutex_unlock(&instance->mutex_filecame);
	
	state = STATE_STARTING;
	
	if (current_media->status & STATUS_HAVE_VIDEO ||
		current_media->status & STATUS_HAVE_AUDIO)
		start_play_this(file_in_cache->filename);
	if (current_media->status & STATUS_HAVE_IMAGE)
		start_show_this(file_in_cache->filename);
	
	return 0;
}



#define PARLEN 127

int player::start_show_this(char *filename)
{
	char tmp[128];
#ifdef DEBUG
	printf("%40s: Starting.\n", __PRETTY_FUNCTION__);
#endif
	
	pthread_mutex_lock(&instance->mutex_end_of_show);
	state |= STATE_WAITING;
	pthread_mutex_unlock(&instance->mutex_end_of_show);

	argc = 0;
	
	argv[argc++] = strdup("display");
	if (window){
		argv[argc++] = strdup("-window");
		snprintf(tmp, PARLEN, "%#x", (int)x_window);
		argv[argc++] = strdup(tmp);
		
		argv[argc++] = strdup("-geometry");
		snprintf(tmp, PARLEN, "%ux%u", 
				window->width, window->height);
		argv[argc++] = strdup(tmp);
	}
	argv[argc++] = strdup(filename);
	argv[argc] = NULL;
	
	pthread_attr_init(&player_thread_attr);
	pthread_attr_setdetachstate(&player_thread_attr,
			     PTHREAD_CREATE_DETACHED);
	pthread_create(&player_thread,
			&player_thread_attr, 
			show_thread_entry,
			this);
	return 0;
}
	
		
int player::start_play_this(char *filename)
{
	char tmp[128];
	int screen;
	
#ifdef DEBUG
	printf("%40s: Starting...\n", __PRETTY_FUNCTION__);
	fflush(stdout);
#endif

	argc = 0;
	
	argv[argc++] = strdup("mplayer");
	
	if (window){
		argv[argc++] = strdup("-double");
		argv[argc++] = strdup("-slave");
	        argv[argc++] = strdup("-nograbpointer");
	        argv[argc++] = strdup("-skin");
	        argv[argc++] = strdup("Blue");

		argv[argc++] = strdup("-wid");
		snprintf(tmp, PARLEN, "%#x", (int)x_window);
		argv[argc++] = strdup(tmp);
		
		argv[argc++] = strdup("-x");
		snprintf(tmp, PARLEN, "%u", window->width);
		argv[argc++] = strdup(tmp);
		
		argv[argc++] = strdup("-y");
		snprintf(tmp, PARLEN, "%u", window->height);
		argv[argc++] = strdup(tmp);

		if (current_media->status &STATUS_HAVE_SPEED){
			argv[argc++] = strdup("-speed");
			snprintf(tmp, PARLEN, "%1.1f", 
					current_media->speed);
			argv[argc++] = strdup(tmp);
		}
		
		if (current_media->status &STATUS_HAVE_BEGIN){
			argv[argc++] = strdup("-ss");
			snprintf(tmp, PARLEN, "%d", current_media->begin);
			argv[argc++] = strdup(tmp);
		}

	}
	argv[argc++] = strdup(filename);
	argv[argc] = NULL;

	/*
	 * Starting a new player thread.
	 */
	pthread_attr_init(&player_thread_attr);
	pthread_attr_setdetachstate(&player_thread_attr,
			     PTHREAD_CREATE_DETACHED);
	pthread_create(&player_thread,
			&player_thread_attr, 
			player_thread_entry,
			this);
	return 0;
}


/*
 * The entry point of the player thread. This will play the already 
 * downloaded file by the already built argv[] array. After the play
 * it returns with NULL (I think this is how a thread should be die).
 */ 
void *player_thread_entry(void *data){
	player *p = (player *)data;
	
	return p->player_thread_function(data);
}


void *player::player_thread_function(void *data)
{
	int input[2];
	int output[2];
	FILE *player_out;
	char *line = NULL;
	size_t nchars = 0;
	file *current_file = NULL;

	printf("%s(): Starting '%s'\n", __PRETTY_FUNCTION__, argv[0]);

	/*
	 * the waiting
	 */
	pthread_mutex_lock(&instance->mutex_filecame);
	current_file = instance->filecache->findfile(current_media->src);
	while (!current_file ||
	       !(current_file->
		 state & (STATE_DOWNLOADED | STATE_FROM_CACHE))) {
		/*
		 * If the file is not yet available, we go to sleep 
		 * until a new file is finished to download. We can 
		 * start the playback only when the whole file is 
		 * downloaded, since mplayer can not seek in truncated
		 * file.
		 */
		state = STATE_WAITING;
		pthread_cond_wait(&instance->cond_filecame, 
				&instance->mutex_filecame);
		current_file = instance->filecache->findfile(current_media->src);
	}
	pthread_mutex_unlock(&instance->mutex_filecame);
	state = STATE_STARTING;

	/*
	 * This pipe will be used to communicate with the child.
	 */
	pipe(input);
	pipe(output);

	pthread_mutex_lock(&instance->mutex_end_of_show);
	state |= STATE_PLAYING;
	pthread_mutex_unlock(&instance->mutex_end_of_show);

	if ( (child=fork()) == 0) {
		/* Child process */
		/* Redirections  */
		dup2(output[0], 0);
		dup2(input[1], 1);
		dup2(input[1], 2);
		close(output[1]);
		close(input[0]);

		execvp(argv[0], (char *const *) argv);
	} else {
		state = STATE_PLAYING;
		
		/* Parent process */
		control_stream = fdopen(output[1], "w");
		close(output[0]);
		close(input[1]);
		player_out = fdopen(input[0], "r");
		while (getdelim(&line, &nchars, '\r', player_out) != -1) {
			if (strlen(line) > 25)
				line[25] = '\0';
			execute_area_tags(line);
		}
		if (state & STATE_QUIT)
			return NULL;

		/* End of the show. */
		control_stream = NULL;

  		while (waitpid(WAIT_ANY, NULL, WNOHANG) > 0)
    		;
#ifdef DEBUG
		printf("%s(): Signaling xml executor.\n", __PRETTY_FUNCTION__);
#endif
		pthread_mutex_lock(&instance->mutex_end_of_show);	
		state = STATE_FREE;
		pthread_cond_signal(&instance->cond_end_of_show);
		pthread_mutex_unlock(&instance->mutex_end_of_show);	
	}

	return NULL;
}


void *show_thread_entry(void *data){
	player *p = (player *)data;
	
	return p->show_thread_function(data);
}

void *player::show_thread_function(void *data)
{
	file *current_file;
	char tmp[32];
	double t;
	
	/* FIXME: This should be a part of the filecache class! */
	pthread_mutex_lock(&instance->mutex_filecame);
	current_file = instance->filecache->findfile(current_media->src);
	while (!current_file ||
	       !(current_file->
		 state & (STATE_DOWNLOADED | STATE_FROM_CACHE))) {
		state = STATE_WAITING;
		pthread_cond_wait(&instance->cond_filecame, 
				&instance->mutex_filecame);
		current_file =
		    instance->filecache->findfile(current_media->src);
	}
	pthread_mutex_unlock(&instance->mutex_filecame);
	state = STATE_STARTING;

	pthread_mutex_lock(&instance->mutex_end_of_show);	
	state |= STATE_PLAYING;
	pthread_mutex_unlock(&instance->mutex_end_of_show);	

	if ((child=fork()) == 0) {
#ifdef DEBUG
		printf("%40s(): Starting '%s'\n", __PRETTY_FUNCTION__,
				argv[0]);
#endif
		execvp(argv[0], (char *const *) argv);
	} else {
		state = STATE_PLAYING;
		t = 0;
		
		while (current_media->status & STATUS_HAVE_DUR &&
				t < current_media->dur){
			usleep(100000);
			//sleep(1);
			t += 0.1;
			sprintf(tmp, "A: %f", t);
			printf("%f   %m     \r", t);
			execute_area_tags(tmp);
			if (state & STATE_QUIT)
				return NULL;
		}
		/* 
		 * End of the show...
		 */
#ifdef DEBUG
		printf("%40s(): Stopping '%d'\n", __PRETTY_FUNCTION__,
				child);
#endif
		kill(child, 1);
		child = -1;

		
  		while (waitpid(WAIT_ANY, NULL, WNOHANG) > 0)
    		;
#ifdef DEBUG
		printf("%40s(): Signaling executor\n", __PRETTY_FUNCTION__);
		fflush(stdout);
#endif
		pthread_mutex_lock(&instance->mutex_end_of_show);	
		state = STATE_FREE;
		pthread_cond_signal(&instance->cond_end_of_show);
		pthread_mutex_unlock(&instance->mutex_end_of_show);	
#ifdef DEBUG
		printf("%40s(): Done\n", __PRETTY_FUNCTION__);
		fflush(stdout);
#endif
	}
	return NULL;
}


/*
 * We return -1 if an error occured, so the player thread does not need to
 * call us again.
 */
int player::execute_area_tags(const char *time_descriptor){
	
	int second;
	area *bigger_area = NULL;
	int n;

	/* If we simply play a file, we can return now. */
	if (!current_media || current_media->nareas == 0)
		return -1;
		
	second = atoi(time_descriptor + 2);
	if (second == current_second)
		return 0;
	
	current_second = second;

	for (n = 0; n < current_media->nareas; ++n)
		if (current_media->areas[n].begin <= second &&
				(!bigger_area ||
				 bigger_area->begin < current_media->areas[n].begin))
			bigger_area = &(current_media->areas[n]);

	if (!bigger_area)
		return 0; /* strange... can it be? */

	if (bigger_area == current_area)
		return 0;

	/*
	 * Now we found a new area needed to activate.
	 */
	current_area = bigger_area;
	NPN_GetURL(instance->instance, current_area->src, current_area->target);
	return 0;
}

void player::set_window(NPWindow *thiswindow)
{
	XGCValues val;

	if (!thiswindow){
		printf("%40s: But this window is NULL.\n", 
				__PRETTY_FUNCTION__);
		return;
	}

	x_ws_info = (NPSetWindowCallbackStruct *) thiswindow->ws_info;
	x_display = x_ws_info->display;

	val.foreground = BlackPixel(x_display, DefaultScreen(x_display));
	val.background = WhitePixel(x_display, DefaultScreen(x_display));
	x_gc = XCreateGC(x_display, (Drawable) thiswindow->window, 
			GCForeground, &val);
	
	val.foreground = WhitePixel(x_display, DefaultScreen(x_display));
	val.background = BlackPixel(x_display, DefaultScreen(x_display));
	x_inv_gc = XCreateGC(x_display, (Drawable) thiswindow->window, 
			GCForeground, &val);
	
	x_font = XQueryFont(x_display, XGContextFromGC(x_gc));
	x_fontheight = x_font->max_bounds.ascent + 
		x_font->max_bounds.descent + 2;

	x_window = (Window)thiswindow->window;	


	window = thiswindow;
}

