/*
 * File: smilplugin.cpp 
 *
 * Program file for smilplugin a SMIL plugin for mozilla. This file
 * is distributed under the terms of GNU Public License (GPL). 
 * (c) 2004 Laszlo Pere (pipas@linux.pte.hu).
 */

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include "smilplugin.h"

#include "filecache.h"
#include "xml.h"
#include "player.h"

#undef DEBUG

NPError NS_PluginGetValue(NPPVariable variable, void *value)
{
#ifdef DEBUG
	printf("%40s(): Browser question.\n", __PRETTY_FUNCTION__);
#endif

	if (variable == NPPVpluginNameString) {
		*((const char **) value) = PACKAGE_STRING;
		return NPERR_NO_ERROR;
	}

	if (variable == NPPVpluginDescriptionString) {
		*((const char **) value) = "Simple SMIL plugin";
		return NPERR_NO_ERROR;
	}

	return NPERR_GENERIC_ERROR;
}

char *NPP_GetMIMEDescription(void)
{
#ifdef DEBUG
	printf("%40s(): Starting.\n", __PRETTY_FUNCTION__);
#endif
	return ("application/smil:smil:"
		"Synchronized Multimedia Integration Language;");
}


void New(nsPluginInstance * instance, nsPluginCreateData * parameters)
{
#ifdef DEBUG
	printf("%40s(): Initializing plugin.\n", __PRETTY_FUNCTION__);
#endif
}

/*
 * The creation and deletion of the plugin instance. The browser will
 * call these functions when it wants a plugin instance of our and
 * when it want to get rid of it.
 */
NPError NS_PluginInitialize()
{
#ifdef DEBUG
	printf("%40s(): Initializing plugin.\n", __PRETTY_FUNCTION__);
#endif

	return NPERR_NO_ERROR;
}

void NS_DestroyPluginInstance(nsPluginInstanceBase * instance)
{
	if (instance)
		delete(nsPluginInstance *) instance;
}

void NS_PluginShutdown(void)
{
	/*
	 * FIXME: it would be a nice step to shut down the player
	 * as fast as we can. Maybe this is the function we need.
	 */
}


nsPluginInstanceBase *NS_NewPluginInstance(nsPluginCreateData
					   * aCreateDataStruct)
{
	if (!aCreateDataStruct)
		return NULL;

	nsPluginInstance *plugin =
	    new nsPluginInstance(aCreateDataStruct->instance);

	return plugin;
}

/***************************************************************************
 * Object: nsPluginInstance                                                *
 *                                                                         *
 *                                                                         *
 *                                                                         *
 *                                                                         *
 ***************************************************************************/
nsPluginInstance::nsPluginInstance(NPP aInstance):nsPluginInstanceBase()
{
#ifdef DEBUG
	printf("%40s(): Creating a new instance.\n", __PRETTY_FUNCTION__);
#endif
	state = STATE_INITIALIZING;
	instance = aInstance;
	window = NULL;
	filecache = NULL;
	myplayer = NULL;
	base_url = NULL;
	pthread_mutex_init(&mutex_filecame, NULL);
	pthread_mutex_init(&mutex_end_of_show, NULL);
	pthread_cond_init(&cond_filecame, NULL);
	pthread_cond_init(&cond_end_of_show, NULL);
}


nsPluginInstance::~nsPluginInstance()
{
#ifdef DEBUG
	printf("%40s(): Destruction.\n", __PRETTY_FUNCTION__);
#endif
	if (myplayer)
		delete myplayer;
	if (xml)
		delete xml;
	if (filecache)
		delete filecache;
	
	pthread_mutex_destroy(&mutex_filecame);
	pthread_mutex_destroy(&mutex_end_of_show);
	pthread_cond_destroy(&cond_filecame);
	pthread_cond_destroy(&cond_end_of_show);
}

NPBool nsPluginInstance::init(NPWindow * thiswindow)
{
	window = thiswindow;
	return TRUE;
}


void nsPluginInstance::shut(void)
{
#ifdef DEBUG
	printf("%40s(): Shutting down plugin.\n", __PRETTY_FUNCTION__);
#endif
	return;
}

NPBool nsPluginInstance::isInitialized()
{
	return TRUE;
}

NPError nsPluginInstance::SetWindow(NPWindow * thiswindow)
{
#ifdef DEBUG
	printf("%40s(): Window size: h=%d w=%d\n",
	       __PRETTY_FUNCTION__, thiswindow->height, thiswindow->width);
#endif
	window = thiswindow;
	return NPERR_NO_ERROR;
}


NPError nsPluginInstance::NewStream(NPMIMEType type, NPStream * stream,
				    NPBool seekable, uint16 * stype)
{
	int cached;
	char *tmp;

#ifdef DEBUG
	printf("%40s(): Starting...\n", __PRETTY_FUNCTION__);
	printf("%40s(): MIME: '%s'\n", __PRETTY_FUNCTION__, type);
	printf("%40s(): Size: %d\n", __PRETTY_FUNCTION__, stream->end);
#endif

	if (!base_url) {
		base_url = strdup(stream->url);
		tmp = base_url + strlen(base_url) - 1;
		while (*tmp != '\0' && *tmp != '/') {
			*tmp = '\0';
			--tmp;
		}
	}

	/* If this is the first file we get, we have to create a
	 * new filecache.
	 */
	if (!filecache)
		filecache = new cachedfiles(this, base_url);

	cached =
	    filecache->newfile(stream->url, (const char *) type,
			       stream->end);

	return cached == 0;
}

NPError nsPluginInstance::DestroyStream(NPStream * stream, NPError reason)
{
#ifdef DEBUG
	printf("%40s(): Starting...\n", __PRETTY_FUNCTION__);
#endif

	return NPERR_NO_ERROR;
}


int32 nsPluginInstance::WriteReady(NPStream * stream)
{
#ifdef DEBUG
	printf("%40s(): Starting...\n", __PRETTY_FUNCTION__);
#endif

	return 4096 * 1024;
}

int32 nsPluginInstance::Write(NPStream * stream, int32 offset,
			      int32 length, void *buffer)
{
	char *lines[32];
	char line[128];
	int32 written;
	file *thisfile;
	const char *tmp;
	int n;

#ifdef DEBUG
	printf("%40s(): Getting data: url='%s'\n", __PRETTY_FUNCTION__,
	       stream->url);
#endif

	if (!filecache)
		filecache = new cachedfiles(this, stream->url);

	written =
	    filecache->writefile(offset, length, buffer, stream->url);

	if (written != length)
		printf("%40s(): Error writing the file: %d\n",
		       __PRETTY_FUNCTION__, written);
#ifdef DEBUG
	filecache->print_cached_file_list();
	printf("%40s(): About to return %d.\n", __PRETTY_FUNCTION__,
	       written);
#endif

	if (myplayer){
		lines[0] = "Downloading...";
		n = 0;
		thisfile = filecache->get_first_file();
		while (thisfile){
			if (strlen(thisfile->url) > 15)
				tmp = thisfile->url + strlen(thisfile->url) -
					15;
			else
				tmp = thisfile->url;
			
			sprintf(line, "%15s: %2.2f%% (%3.1fM)", tmp, 
					(double)thisfile->actualsize /
					(double)thisfile->filesize *
					100, 
					(double)thisfile->actualsize /
					(1024*1024) );
			lines[n] = strdupa(line);
			thisfile = filecache->get_next_file();
			n++;
			if (n >= 31)
				break;
		}
		lines[n] = NULL;
		myplayer->draw_status(lines);
	}
	
	return written;
}


/*
 * This function must be rewritten befora public any release!
 */
int nsPluginInstance::FileDownloaded(char *filename, char *mimetype)
{
#ifdef DEBUG
	printf("%40s(): Starting.\n", __PRETTY_FUNCTION__);
#endif

	if (!myplayer) {
		myplayer = new player(this);
		myplayer->set_window(window);
	}

	if (mimetype && strcasecmp(mimetype, "application/smil") == 0) {
		xml = new xml_parser(this, filename);
		return 0;
	}
	
	if(!(state & STATE_EXECUTED)){
		state |= STATE_EXECUTED;
		xml->execute_document();
	}

	/*
	 * We signal every thread waiting for a file, maybe some of them
	 * are lucky and the file came they are waiting for.
	 */
	pthread_mutex_lock(&mutex_filecame);	
	pthread_cond_broadcast(&cond_filecame);	
	pthread_mutex_unlock(&mutex_filecame);	
	return 0;
}


int nsPluginInstance::print_message(char *message)
{
	char *lines[32];

	printf("message: %s\n", message);

	if (!myplayer)
		return -1;
	
	lines[0] = message;
	lines[1] = NULL;
	myplayer->draw_status(lines);
	return 0;
}

