#include "xml.h"
#include <unistd.h>
#include "player.h"

#undef DEBUG

xml_parser::xml_parser(nsPluginInstance * plugin_instance,
		       char *xmlfilename)
{
	instance = plugin_instance;
	smil_doc = NULL;
	filename = NULL;

	if (xmlfilename) {
		filename = strdup(xmlfilename);
		parse_document(filename);
	}
}

xml_parser::~xml_parser(void)
{
	if (smil_doc)
		xmlFreeDoc(smil_doc);
	if (filename)
		free(filename);
}


/* 
 * This is an important function. It will execute (play) the 
 * content of an xml smil file by walking through the memory 
 * representation.
 */
int xml_parser::execute_document(void)
{
	if (!smil_doc) {
		printf("%40s(): The smil_document is empty.\n",
		       __PRETTY_FUNCTION__);
		return -1;
	}
	root = xmlDocGetRootElement(smil_doc);
	if (!root) {
		printf("%40s(): This smil document seems to be empty.\n",
		       __PRETTY_FUNCTION__);
		return -1;
	}

	/*
	 * We create a new thread toplay the given tree.
	 */
	pthread_attr_init(&player_thread_attr);
	pthread_attr_setdetachstate(&player_thread_attr,
				    PTHREAD_CREATE_DETACHED);
	pthread_create(&player_thread,
		       &player_thread_attr, execute_thread_entry, this);

	return 0;
}

void *execute_thread_entry(void *data)
{
	xml_parser *p = (xml_parser *) data;
	p->do_execute(p->root);
	return NULL;
}


/*
 * This recursive function used to walk through the xml tree
 * and execute (play) its contents.
 */
int xml_parser::do_execute(xmlNodePtr current)
{
	if (current->xmlChildrenNode)
		if (do_execute(current->xmlChildrenNode) != 0) {
			printf("%40s(): Interrupting execution.\n",
			       __PRETTY_FUNCTION__);
			return -1;
		}

	while (current != NULL) {
#ifdef DEBUG
	if (current->name)
		printf("%40s: Executing a '%s'.\n", __PRETTY_FUNCTION__,
		       current->name);
	if (current->ns){
		if (current->ns->href)
			printf("%40s: Namespace href: '%s'\n", __PRETTY_FUNCTION__,
			current->ns->href);
		if (current->ns->prefix)
			printf("%40s: Namespace prefix: '%s'\n", __PRETTY_FUNCTION__,
			current->ns->prefix);
	}

	
#endif
		if (current->content) {
#ifdef DEBUG
			printf
			    ("%40s(): Execution of content found at %p.\n",
			     __PRETTY_FUNCTION__, current->content);
#endif

			if (!instance || !instance->myplayer) {
				printf("%40s(): Player not found.\n",
				       __PRETTY_FUNCTION__);
				return -1;
			}

			pthread_mutex_lock(&instance->mutex_end_of_show);
			if (instance->myplayer->state) {
				pthread_cond_wait(&instance->
						  cond_end_of_show,
						  &instance->
						  mutex_end_of_show);
			}

			pthread_mutex_unlock(&instance->mutex_end_of_show);

			/*
			 * Now we start the playback in a new thread...
			 */
			instance->myplayer->
			    start_play_this((media_instance_descriptor *)
					    current->content);
		}
		current = current->next;
	}
#ifdef DEBUG
	printf("%40s(): No more work to do.\n", __PRETTY_FUNCTION__);
	fflush(stdout);
#endif
	return 0;
}



void xml_parser::parse_areas(xmlNodePtr area,
			     media_instance_descriptor * media_descriptor)
{
	char *src = NULL;
	char *target = NULL;
	char *begin = NULL;
	char *end = NULL;
	char *tmp;

	xmlAttrPtr current_property;

	while (area != NULL) {
		current_property = area->properties;
		while (current_property) {
			tmp =
			    (char *) xmlNodeListGetString(smil_doc,
							  current_property->
							  val, 1024);
			if (xmlStrcmp
			    (current_property->name,
			     (const xmlChar *) "src") == 0
			    || xmlStrcmp(current_property->name,
					 (const xmlChar *) "href") == 0)
				src = tmp;
			if (xmlStrcmp
			    (current_property->name,
			     (const xmlChar *) "target") == 0)
				target = tmp;
			if (xmlStrcmp
			    (current_property->name,
			     (const xmlChar *) "begin") == 0)
				begin = tmp;
			if (xmlStrcmp
			    (current_property->name,
			     (const xmlChar *) "end") == 0)
				end = tmp;

			current_property = current_property->next;
		}

		/* Store data to the media desriptor is a simple task. :-)
		 */
		if (target && begin)
			media_descriptor->insert_area(src, target, begin,
						      end);

		if (src)
			xmlFree(src);
		if (target)
			xmlFree(target);
		if (begin)
			xmlFree(begin);
		if (end)
			xmlFree(end);
		src = NULL;
		target = NULL;
		begin = NULL;
		end = NULL;
		area = area->next;
	}
}


void xml_parser::parse_media(xmlNodePtr media)
{
	xmlNodePtr current;
	xmlAttrPtr current_property;
	xmlChar *value = NULL;
	char *tmp;

#ifdef DEBUG
	printf("%40s: Media name: '%s'\n", __PRETTY_FUNCTION__,
	       media->name);
#endif
	media_instance_descriptor *Media = new media_instance_descriptor();

	if (!xmlStrcmp(media->name, (const xmlChar *) "video"))
		Media->status |= STATUS_HAVE_VIDEO;

	if (!xmlStrcmp(media->name, (const xmlChar *) "audio"))
		Media->status |= STATUS_HAVE_AUDIO;

	if (!xmlStrcmp(media->name, (const xmlChar *) "img"))
		Media->status |= STATUS_HAVE_IMAGE;

	current_property = media->properties;
	while (current_property) {
		tmp =
		    (char *) xmlNodeListGetString(smil_doc,
						  current_property->val,
						  1024);

		if (xmlStrcmp
		    (current_property->name, (const xmlChar *) "src") == 0)
			Media->src = tmp;

		if (xmlStrcmp
		    (current_property->name, (const xmlChar *) "id") == 0)
			Media->id = tmp;

		if (xmlStrcmp
		    (current_property->name,
		     (const xmlChar *) "begin") == 0) {
			Media->begin = atoi(tmp);
			Media->status |= STATUS_HAVE_BEGIN;
			xmlFree(tmp);
		}

		if (xmlStrcmp
		    (current_property->name,
		     (const xmlChar *) "end") == 0) {
			Media->begin = atoi(tmp);
			xmlFree(tmp);
		}

		if (xmlStrcmp
		    (current_property->name,
		     (const xmlChar *) "dur") == 0) {
			Media->dur = atoi(tmp);
			Media->status |= STATUS_HAVE_DUR;
			xmlFree(tmp);
		}
		
		if (xmlStrcmp
		    (current_property->name,
		     (const xmlChar *) "speed") == 0) {
			Media->speed = atof(tmp);
			Media->status |= STATUS_HAVE_SPEED;
			xmlFree(tmp);
		}

		if (xmlStrcmp
		    (current_property->name,
		     (const xmlChar *) "repeatCount") == 0) {
			Media->repeatCount = atoi(tmp);
			xmlFree(tmp);
		}

		current_property = current_property->next;
	}

	/*
	 * We check if we have child node. Those must be area nodes descriptors of the 
	 * small areas of media. They are defining what to do when we playing the 
	 * given portion of the media.
	 */
	parse_areas(media->xmlChildrenNode, Media);

	/*
	 * I'm not sure I can do this, but it is so simple just store the data
	 * in the node and walk through the tree again...
	 */
	media->content = (xmlChar *) Media;

	/* FIXME: This is just a temporary shortcut. Things are much more complicated.
	 */
	if (Media->src)
		NPN_GetURL(instance->instance, (const char *) Media->src,
			   NULL);

	return;
}

void xml_parser::parse_seq(xmlNodePtr seq)
{
	xmlNodePtr current;

	current = seq->xmlChildrenNode;
	while (current != NULL) {
		printf("%40s(): Children of seq '%s'.\n",
		       __PRETTY_FUNCTION__, current->name);

		if (!xmlStrcmp(current->name, (const xmlChar *) "video"))
			parse_media(current);

		if (!xmlStrcmp(current->name, (const xmlChar *) "audio"))
			parse_media(current);

		if (!xmlStrcmp(current->name, (const xmlChar *) "img"))
			parse_media(current);

		current = current->next;
	}

	return;
}

void xml_parser::parse_body(xmlNodePtr body)
{
	xmlNodePtr current;

	current = body->xmlChildrenNode;
	while (current != NULL) {
		if (xmlStrcmp(current->name, (const xmlChar *) "seq") == 0)
			parse_seq(current);
		/*
		 * FIXME: these tags are missing, so I try to handle them as
		 * seq tags.
		 */
		if (xmlStrcmp(current->name, (const xmlChar *) "par") == 0)
			parse_seq(current);
		if (xmlStrcmp(current->name, (const xmlChar *) "excl") == 0)
			parse_seq(current);

		current = current->next;
	}

	return;
}

/*
 * FIXME: This function should kill the object when ready.
 */
void xml_parser::parse_document(char *filename)
{
	xmlNodePtr cur;
	int n = 0;

	smil_doc = xmlParseFile(filename);

	if (!smil_doc) {
		printf("%40s(): Document not parsed successfully.\n",
		       __PRETTY_FUNCTION__);
		if (instance)
			instance->print_message("SMIL parse error.");
	}

	cur = xmlDocGetRootElement(smil_doc);

	if (!cur) {
		printf("%40s(): Empty document\n", __PRETTY_FUNCTION__);
		xmlFreeDoc(smil_doc);
		return;
	}

	if (xmlStrcmp(cur->name, (const xmlChar *) "smil")) {
		printf("%40s(): Document of the wrong type.\n",
		       __PRETTY_FUNCTION__);
		xmlFreeDoc(smil_doc);
		return;
	}

	cur = cur->xmlChildrenNode;
	while (cur != NULL) {
		printf("%40s(): Children '%s'.\n", __PRETTY_FUNCTION__,
		       cur->name);
		if (xmlStrcmp(cur->name, (const xmlChar *) "body") == 0)
			break;
		cur = cur->next;
	}

	if (xmlStrcmp(cur->name, (const xmlChar *) "body") != 0)
		printf("%40s(): No body found.\n", __PRETTY_FUNCTION__);
	else
		parse_body(cur);

	return;
}

/***********************************************************************
 * Class: media                                                        *
 *                                                                     *
 *                                                                     *
 *                                                                     *
 *                                                                     *
 *                                                                     *
 ***********************************************************************/

media_instance_descriptor::media_instance_descriptor(void)
{
	src = NULL;
	id = NULL;

	areas = (area *) NPN_MemAlloc(128 * sizeof(area));
	nareas = 0;
	maxareas = 128;

	begin = 0;
	end = 0;
	dur = 0;
	repeatCount = 0;

	status = STATUS_EMPTY;
}

media_instance_descriptor::~media_instance_descriptor()
{
	if (src)
		xmlFree(src);
	if (id)
		xmlFree(id);

	/* FIXME: we have to free the string fields in area space.
	 */
}

void media_instance_descriptor::print_debug(void)
{
	int n;

	if (src)
		printf("%40s(): src='%s'\n", __PRETTY_FUNCTION__, src);
	else
		printf("%40s(): src=(unknown)\n", __PRETTY_FUNCTION__);

	if (id)
		printf("%40s(): id='%s'\n", __PRETTY_FUNCTION__, id);
	else
		printf("%40s(): id=(unknown)\n", __PRETTY_FUNCTION__, id);

	printf("%40s(): begin=%u\n", __PRETTY_FUNCTION__, begin);
	printf("%40s(): end=%u\n", __PRETTY_FUNCTION__, end);
	printf("%40s(): dur=%u\n", __PRETTY_FUNCTION__, dur);
	printf("%40s(): repeatCount=%u\n", __PRETTY_FUNCTION__,
	       repeatCount);

	/*
	 * I want to see if the area tags are ok.
	 */
	for (n = 0; n < nareas; ++n)
		printf
		    ("%40s: begin: %d, end: %d, target: '%s' src: '%s'\n",
		     __PRETTY_FUNCTION__, areas[n].begin, areas[n].end,
		     areas[n].target, areas[n].src);
}


int media_instance_descriptor::insert_area(char *src, char *target,
					   char *begin, char *end)
{

	if (!areas || nareas == maxareas) {
		printf("%40s(): No space for areas.\n",
		       __PRETTY_FUNCTION__);
		return -1;
	}

	if (src)
		areas[nareas].src = strdup(src);
	else
		areas[nareas].src = strdup("unknown");

	if (target)
		areas[nareas].target = strdup(target);
	else
		areas[nareas].target = strdup("unknown");

	if (begin)
		areas[nareas].begin = atoi(begin);
	else
		areas[nareas].begin = 0;

	if (end)
		areas[nareas].end = atoi(end);
	else
		areas[nareas].end = 0;

	++nareas;
}
